/* ======================================================================
 *
 *  CCCCC    ++      ++   pfTravel.c++
 * CC   CC   ++      ++   Author(s): Bill Sherman
 * CC      ++++++  ++++++ Created: March 1, 2000
 * CC   CC   ++      ++   Last Modified: February 12, 2004 -- Happy Birthday Abe!
 *  CCCCC    ++      ++
 *
 * =================================================================== */
/***********************************************************************/
/* This example code demonstrates a Performer application program with */
/*   a travel interface controlled by the joystick.                    */
/*                                                                     */
/* Copyright 2004, William R. Sherman, All Rights Reserved.            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

#define PF_YUP	/* undefine to test new internal Z-up coordinate system */

/* Performer library includes */
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfDCS.h>		/* also handles pfSCS */
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pfdu.h>
#include <Performer/prmath.h>		/* for pfMatrix & pfVec4 */

/* VR library includes */
#define GFX_PERFORMER
#include "freevr.h"
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */


/* Define a structure for information about each individual object. */
typedef struct {
		pfDCS		*dcs;		/* position of the object (controlled by a pfDCS) */
		pfVec4		*color;		/* vector that contains values used to color the object */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		object_count;
		int		selected_object;
		ObjectType	object[5];

		pfDCS		*wand_dcs;	/* pointer to the DCS the positions the wand avatar */
		pfVec4		*wand_color;
	} WorldDataType;

/* This is very important -- this function will be used as an argument. */
	void	exit_application(int);

/* functions defined in this file */
	void	update_world(WorldDataType *wd);
static	pfScene	*createScene(WorldDataType *wd);
static	int	world_travel(pfTraverser *trav, void *);


int main(int argc, char **argv)
{
	WorldDataType	*wd;

	/* TODO: I do the signal setup a little different in Pf than OpenGL -- investigate */
	signal(SIGINT, SIG_IGN);

	/* initialize Performer */
	pfInit();

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, exit_application);
	vrConfigure(&argc, argv, NULL);

	/* configure Performer */
	pfConfig();

	/* create the channels for viewing the world */
	vrStart();

	/***************************************************************/
	/* Level of detail of performance information is higher for    */
	/*   simulator display.  This allows greater debugging ability */
	/*   at the expense of some performance.                       */
#if 0 /* this causes core-dump on Linux-Performer 2.3 */
	/* Turn on all possible statistics, except "fx pipe fill" */
	vrPfMasterChannel()->getFStats()->setClass(PFSTATS_ALL ^ PFSTATSHW_ENGFXPIPE_FILL, PFSTATS_ON);
#endif


	/********************/
	/* create the world */
	wd = (WorldDataType *)pfMalloc(sizeof(WorldDataType), pfGetSharedArena());

	/* Create a scene and assign it to the channels */
	vrPfMasterChannel()->setScene(createScene(wd));


	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* do the main simulation/render loop until 'Esc' key is pressed */
	while (!vrGet2switchValue(0)) {

		/* In the non-Peformer model, most normal world operations */
		/* would go here -- in update_world().  With Performer,    */
		/* applications can also use callback routines within the  */
		/* scene-graph to modify the world.                        */
		update_world(wd);


		pfSync();
		/* latency critical operations go here */
		vrPfPreFrame();
		pfFrame();
		vrPfPostFrame();
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int)
{
	vrExit();
	pfExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine.                      */
void update_world(WorldDataType *wd)
{
#define	UPDATE_WAND_BYHAND	/* define this to calculate the wand avatar's position here */

#ifdef UPDATE_WAND_BYHAND
	float		wand_orient[3];
	vrEuler		wand_euler;
#endif
static	vrMatrix	*mat_neg90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X, -90.0);	/* convert from Pf to OpenGL space */
static	vrMatrix	*mat_pos90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X,  90.0);	/* convert from OpenGL to Pf space  (ogl2pf) */
	vrPoint		wand_rw_point;
	vrPoint		wand_point;
#define wand_rw_location wand_rw_point.v
#define wand_location wand_point.v
	pfMatrix	obj_mat;
	float		obj_x, obj_y, obj_z, tmp;
	ObjectType	*object;
	int		prev_selected_object;
	int		obj;
	int		grab_button;			/* will be 1 when button is held down */
	pfMatrix	tmpmat1, tmpmat2;		/* These are used by direct method of wand avatar positioning. */

#if 0 /* static world for now */
	float		sim_time = pfGetTime();

	obj_z = 5.5;
	obj_x =  3*sin(sim_time * 2.0);
	obj_y = -3*cos(sim_time * 2.0) + 4.0;
	wd->object[0].dcs->setTrans(obj_x, obj_y, obj_z);

	obj_y = 4.0;
	obj_x = 3*cos(sim_time * 2.0);
	obj_z = 3*sin(sim_time * 2.0) + 5.5;
	wd->object[1].dcs->setTrans(obj_x, obj_y, obj_z);
#endif

	/* need to update the wand's avatar in the sim-process (aka APP process) */
	/*   NOTE: we could also do this with an APP-callback on the DCS node */
#ifndef UPDATE_WAND_BYHAND /* calculate by hand (0), or use VR library routine (1) { */
	/* calculate the wand DCS (in RW) using the VR library */
	vrPfDCSTransform6sensor(wd->wand_dcs, WAND_SENSOR);

	/* now correct the wand orientation */
#  ifdef PF_YUP
	/* (ie. Rotate about X to make wand avatar point forward instead of up) */
	tmpmat1.makeRot(-90.0, 1.0, 0.0, 0.0);
#  else
	/* (ie. Rotate about X to make wand avatar point forward instead of aftward) */
	tmpmat1.makeRot(180.0, 1.0, 0.0, 0.0);
#  endif
	wd->wand_dcs->getMat(tmpmat2);	/* WARNING: unobvious call-by-ref */
	tmpmat2.preMult(tmpmat1);
	wd->wand_dcs->setMat(tmpmat2);	/* WARNING: unobvious call-by-ref */

#  ifdef PF_YUP
	/* NOTE: in older FreeVR, the scale of a pfDCS was not maintained when set by vrPfDCSTransform6sensor() */
	wd->wand_dcs->setScale(0.25, 0.25, 0.25);
#  endif
#else /* } UPDATE_WAND_BYHAND { */
	/* calculate wand DCS (in RW) by hand */
	vrEulerGetRWFrom6sensor(&wand_euler, WAND_SENSOR);
	wand_rw_point.v[VR_X] = wand_euler.t[VR_X];
	wand_rw_point.v[VR_Y] = wand_euler.t[VR_Y];
	wand_rw_point.v[VR_Z] = wand_euler.t[VR_Z];
	vrPointTransformByMatrix(&wand_rw_point, &wand_rw_point, mat_pos90x);	/* adjust to put into Performer space */

	/* convert to CL style orientation, also with regard to Performer coordinate system */
	wand_orient[2] = wand_euler.r[VR_ELEV];
	wand_orient[1] = wand_euler.r[VR_ROLL];	/* would be AZIM in non pf */
	wand_orient[0] = wand_euler.r[VR_AZIM];	/* would be ROLL in non pf */
	/* NOTE: roll is not included here because the cave uses a */
	/*   different system of Eulers (ie. order of rotations)   */
	wd->wand_dcs->setRot(wand_orient[2], wand_orient[0]-90.0, 0.0);
	wd->wand_dcs->setTrans(wand_rw_location[0], wand_rw_location[1], wand_rw_location[2]);	/* TODO: (bug!) This isn't correct, wand_location has VW location, but need to render avatar in RW (also, the rotations seem a bit out of whack) */
	/* [2/13/04: and the problem is we need to render in RW coords, but use VW coords for object selection/manipulation */
#endif /* } */

	/* get the VW location of the wand */
	/* TODO: really this should be based on the concept of props, but nyi */
#  ifdef PF_YUP
	vrPointGetVWFromUser6sensor(&wand_point, 0, WAND_SENSOR);
	vrPointTransformByMatrix(&wand_point, &wand_point, mat_pos90x);		/* adjust to put into Performer space */
#  else
	vrPfPointGetVWFromUser6sensor(&wand_point, 0, WAND_SENSOR);
#  endif

	/* check for contact between the wand and all the objects */
	/* NOTE: this has a slight "quirk," in that the highest numbered */
	/*   object always becomes selected, even when one of the objects*/
	/*   was already selected.  A better UI would not change the     */
	/*   selected object while the button is held down.              */
	prev_selected_object = wd->selected_object;
	wd->selected_object = -1;
	for (obj = 0; obj < wd->object_count; obj++) {
		wd->object[obj].dcs->getMat(obj_mat);
		obj_mat.getRow(3, &obj_x, &obj_y, &obj_z, &tmp);
		if ( (wand_location[VR_X] < obj_x + 1.0)
		  && (wand_location[VR_X] > obj_x - 1.0)
		  && (wand_location[VR_Y] < obj_y + 1.0)
		  && (wand_location[VR_Y] > obj_y - 1.0)
		  && (wand_location[VR_Z] < obj_z + 1.0)
		  && (wand_location[VR_Z] > obj_z - 1.0) ) {
			wd->selected_object = obj;
		}
	}

	/* the selection changes, then make a change to the world to reflect this */
	if (prev_selected_object != wd->selected_object) {
		if (prev_selected_object != -1) {
			/* restore the previosly selected object */
			*(wd->object[prev_selected_object].color) *= 2.0;
		}
		if (wd->selected_object != -1) {
			/* enhance the currently selected object */
			*(wd->object[wd->selected_object].color) *= 0.5;

		}
	}
	prev_selected_object = wd->selected_object;

	grab_button = vrGet2switchValue(1);

	/* when left wand button is pressed, the selected object is moved */
	if (grab_button && (wd->selected_object != -1)) {
		object = &(wd->object[wd->selected_object]);
		/* Note this is a little different than the OpenGL */
		/*   version because the object will jump to be    */
		/*   centered about the wand.                      */
		object->dcs->setTrans(wand_location[0], wand_location[1], wand_location[2]);
	}


	/********************************************/
	/* allow other processes to get in edgewise */
	sginap(0);
}


/**********************************************************************/
static int world_travel(pfTraverser *trav, void *)
{
#define SPEED 0.02f
static	double		last_time = -1.0;
	double		this_time = pfGetTime();
	double		delta_time;
	float		delta_move;
	pfDCS   	*travel_dcs = (pfDCS *)trav->getNode();  /* Get a pointer to the DCS node itself */
	float		joy_x, joy_y;
	vrVector	wand_pointvec;

	/***************************************************/
	/* Determine delta time from last simulation frame */
	if (last_time == -1.0)
		last_time = this_time;

	delta_time = this_time - last_time;

	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	/* notice rotation for Performer is about the z-axis instead of y */
	if (fabs(joy_x) > 0.125)
#ifdef PF_YUP
		vrUserTravelRotateId(VR_ALLUSERS, VR_Y, (joy_x * delta_time * SPEED * -1.50));
#else
		vrUserTravelRotateId(VR_ALLUSERS, VR_Z, (joy_x * delta_time * SPEED * -1.50));
#endif

	if (fabs(joy_y) > 0.125) {
		delta_move = joy_y * delta_time * SPEED;
		vrVectorGetRWFrom6sensorDir(&wand_pointvec, WAND_SENSOR, VRDIR_FORE);
		vrUserTravelTranslate3d(VR_ALLUSERS,
			wand_pointvec.v[VR_X] * delta_move,
			wand_pointvec.v[VR_Y] * delta_move,
			wand_pointvec.v[VR_Z] * delta_move);
	}

	/* Reset travel on button 3 */
	if (vrGet2switchValue(3))
		vrUserTravelReset(VR_ALLUSERS);

	/* Load the DCS with the latest navigation matrix */
 #if 0
	/* NOTE: this function relys on the current user to modify the dcs node.  */
	/*   This is going to be a problem.  (Especially since we are here during */
	/*   the APP process.)  Ideally, a DCS node would be assigned this as a   */
	/*   callback that would occur during the Cull/Draw process.              */
	vrPfDCSTransformUserTravel(travel_dcs);
 #else
	/* NOTE: In FreeVR, because each user can have their own travel CS, the  */
	/*   pfDCS cannot be changed in the APP callback.  So the AP must assign */
	/*   the special callback function to the travel node as a CULL callback.*/
	/*   The function is vrPfDCSTransformUserTravelCB().                     */
 #endif

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

pfNode *make_cube(pfVec4 *color_vec);
pfNode *make_pyramid(pfVec4 *color_vec);

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.).  Then a simple     */
/* world consisting of a cube and pyramid are created.                       */
/*****************************************************************************/
static pfScene *createScene(WorldDataType *wd)
{
	pfScene		*scene;
	pfGeoState	*gstate;
	pfVec4		*color;
	pfDCS		*cs_node;
	pfDCS		*travel_cs;

	wd->object_count = 0;
	wd->selected_object = -1;

	scene = new pfScene;

	/* Set up a base geostate for the scene */
	gstate = new pfGeoState;
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);


	/***************************************/
	/*** Create a DCS for travel control ***/
	travel_cs = new pfDCS();
	scene->addChild(travel_cs);

	/* use a callback to adjust the navigation */
	travel_cs->setTravFuncs(PFTRAV_APP, world_travel, NULL);
#ifdef PF_YUP
	travel_cs->setTravFuncs(PFTRAV_CULL, vrPfDCSTransformUserTravelCB, NULL);
#else
	travel_cs->setTravFuncs(PFTRAV_CULL, vrPfDCSTransformUserTravelCB_zup, NULL);
#endif


	/**********************************************************/
	/*** Create a DCS (w/ matrix) for each *dynamic* object ***/

#if 0
	/* a blue floor (ie. large flat cube) */
	cs_node = new pfDCS();
	travel_cs->addChild(cs_node);
	cs_node->setTrans(0.0, 0.0, 0.0);
	cs_node->setScale(10.0, 10.0, 0.1);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(25.0/256.0, 25.0/256.0, 200.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));
#endif

	/* a blue cube */
	cs_node = new pfDCS();
	travel_cs->addChild(cs_node);
	cs_node->setTrans(-2.0, 4.0, 3.5);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 100.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a red pyramid */
	cs_node = new pfDCS();
	travel_cs->addChild(cs_node);
	cs_node->setTrans( 2.0, 4.0, 3.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(255.0/256.0, 100.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(color));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a green upside down pyramid */
	cs_node = new pfDCS();
	travel_cs->addChild(cs_node);
	cs_node->setTrans( 0.0, 4.0, 4.0);
	cs_node->setScale(1.0, 1.0, -1.0);	/* NOTE: flip in Z, not Y */
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 255.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(color));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* the Wand -- a long cyan pyramid */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	wd->wand_dcs = cs_node;
	cs_node->setScale(0.25, 0.25, 0.25);
	/* NOTE: setting rotation and translation of this DCS will be uneffective   */
	/*   because those values will be overwritten by the Sensor transformation. */
	wd->wand_color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	wd->wand_color->set(100.0/256.0, 255.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(wd->wand_color));

	/*** the rest of the objects are in the RWcs, not the VWcs ***/

	/* a 6' rose cube behind the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(0.0, -10.0, 5.0);
	cs_node->setScale(6.0, 6.0, 6.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(240.0/256.0, 138.0/256.0, 133.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a 4' blue cube in front of the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(0.0, 7.0, 3.0);
	cs_node->setScale(4.0, 4.0, 4.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(0.0/256.0, 0.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a 4' white cube left of the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(-7.0, 0.0, 3.0);
	cs_node->setScale(4.0, 4.0, 4.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(255.0/256.0, 255.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a pink pyramid on the floor of the cave near the front wall */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(0.0, 4.0, 0.0);
	cs_node->setScale(2.0, 2.0, 2.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(255.0/256.0, 100.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(color));

	/* a 2' light purple cube at the front left corner of the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(-6.0, 6.0, 7.0);
	cs_node->setScale(2.0, 2.0, 2.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 100.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a 2' green cube below the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(5.0, 0.0, 1.0);
	cs_node->setScale(2.0, 2.0, 2.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(0.0/256.0, 255.0/256.0, 0.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a 2' dark purple cube below the center of the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(0.0, 0.0, -1.0);
	cs_node->setScale(2.0, 2.0, 2.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 0.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a 10' tall red pillar inside the right front corner of the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(4.5, 4.5, 5.0);
	cs_node->setScale(1.0, 1.0, 10.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(240.0/256.0, 50.0/256.0, 50.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a 10' deep cyan pillar inside the right front corner of the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(4.5, 4.5, -5.0);
	cs_node->setScale(1.0, 1.0, 10.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(50.0/256.0, 240.0/256.0, 240.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a 4' tall yellow pillar near the right wall of the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(-2.5, -2.0, 2.0);
	cs_node->setScale(1.0, 1.0, 4.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(240.0/256.0, 240.0/256.0, 50.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a 4' long orange pillar out of the right wall of the cave */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(-3.0, 2.5, 2.0);
	cs_node->setScale(4.0, 1.0, 1.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(240.0/256.0, 130.0/256.0, 30.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	return (scene);
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_cube(pfVec4 *color_vec)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewCube(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/*******************************************/
/* make a unit pyramid of the given color. */
/*   Note that this pyramid is smaller than the shapes.c cube. */
pfNode *make_pyramid(pfVec4 *color_vec)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewPyramid(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}

