/* ======================================================================
 *
 *  CCCCC          vr_basicgfx.glx.c
 * CC   CC         Author(s): Bill Sherman, John Stone
 * CC              Created: November 30, 1999
 * CC   CC         Last Modified: May 2, 2006
 *  CCCCC
 *
 * Code file for FreeVR basic graphics routines.
 *
 * NOTE: these routines make some assumptions about the current
 *   graphics rendering state.  Namely:
 *	- counterclockwise ordering of vertices designates the front face of a polygon
 *      ??? lighting
 *	??? blending enable/function
 * Also, the outline lines will be rendered at whatever the current linewidth is
 *
 * Copyright 2006, Bill Sherman & Friends, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
#include <GL/gl.h>

#include <stdio.h>
#include "vr_debug.h"
#include "vr_basicgfx.glx.h"
#include "vr_input.h"		/* for input sensing functions in vrGLRenderDefaultSimultor() */


/****************************************************************************/
/** code from Bill's shapes.c                                              **/
/****************************************************************************/

static float cube_vertices[9][3] = {			/* specifies the (x,y,z)   */
		{ -1.0,  1.0,  1.0 },	/* 0 */	/* location of each vertex */
		{ -1.0, -1.0,  1.0 },	/* 1 */
		{  1.0, -1.0,  1.0 },	/* 2 */
		{  1.0,  1.0,  1.0 },	/* 3 */
		{  1.0,  1.0, -1.0 },	/* 4 */
		{  1.0, -1.0, -1.0 },	/* 5 */
		{ -1.0, -1.0, -1.0 },	/* 6 */
		{ -1.0,  1.0, -1.0 },	/* 7 */
		{  0.0,  1.0,  0.0 },	/* 8 */
	};

static int cube_polygons[6][4] = {			/* specifices the vertices */
		{ 0, 1, 2, 3 },	/* front  */	/* of each polygon         */
		{ 3, 2, 5, 4 },	/* right  */
		{ 5, 6, 7, 4 },	/* back   */
		{ 7, 6, 1, 0 },	/* left   */
		{ 6, 5, 2, 1 },	/* bottom */
		{ 0, 3, 4, 7 },	/* top    */
	};

static float cube_normals[6][3] = {			/* specifies the normal */
		{  0.0,  0.0,  1.0 }, /* front  ** of each polygon      */
		{  1.0,  0.0,  0.0 }, /* right  */
		{  0.0,  0.0, -1.0 }, /* back   */
		{ -1.0,  0.0,  0.0 }, /* left   */
		{  0.0, -1.0,  0.0 }, /* bottom */
		{  0.0,  1.0,  0.0 }, /* top    */
	};

static int pyramid_polygons[5][4] = {			/* specifices the vertices */
		{ 1, 2, 8 },	/* front  */	/* of each polygon         */
		{ 2, 5, 8 },	/* right  */
		{ 5, 6, 8 },	/* back   */
		{ 6, 1, 8 },	/* left   */
		{ 6, 5, 2, 1 },	/* bottom */
	};

static float pyramid_normals[5][3] = {			/* specifies the normal    */
		{  0.0,  0.0,  1.0 }, /* front  ** of each polygon -- only */
		{  1.0,  0.0,  0.0 }, /* right  ** the bottom normal is    */
		{  0.0,  0.0, -1.0 }, /* back   ** perpendicular to the    */
		{ -1.0,  0.0,  0.0 }, /* left   ** surface.                */
		{  0.0, -1.0,  0.0 }, /* bottom */
	};


/*********************************************************************/
void vrShapeGLCube()
{
	int	poly;

	for (poly = 0; poly < 6; poly++) {
		glBegin(GL_POLYGON);
			glNormal3fv(cube_normals[poly]);
			glVertex3fv(cube_vertices[cube_polygons[poly][0]]);
			glVertex3fv(cube_vertices[cube_polygons[poly][1]]);
			glVertex3fv(cube_vertices[cube_polygons[poly][2]]);
			glVertex3fv(cube_vertices[cube_polygons[poly][3]]);
		glEnd();
	}
}


/*********************************************************************/
void vrShapeGLCubeOpen()
{
	int	poly;

	for (poly = 0; poly < 6; poly++) {
		if (poly != 0) {
			glBegin(GL_POLYGON);
				glNormal3fv(cube_normals[poly]);
				glVertex3fv(cube_vertices[cube_polygons[poly][0]]);
				glVertex3fv(cube_vertices[cube_polygons[poly][1]]);
				glVertex3fv(cube_vertices[cube_polygons[poly][2]]);
				glVertex3fv(cube_vertices[cube_polygons[poly][3]]);
			glEnd();
		}
	}
}


/*********************************************************************/
void vrShapeGLCubeOutline()
{
	glPushAttrib(GL_COLOR_BUFFER_BIT);
	glDisable(GL_BLEND);
	glBegin(GL_LINE_STRIP);
		glVertex3fv(cube_vertices[0]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[3]);
		glVertex3fv(cube_vertices[0]);
		glVertex3fv(cube_vertices[7]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[7]);
		glVertex3fv(cube_vertices[4]);
		glVertex3fv(cube_vertices[3]);
		glVertex3fv(cube_vertices[4]);
		glVertex3fv(cube_vertices[5]);
	glEnd();
	glPopAttrib();
}


/*********************************************************************/
void vrShapeGLCubePlusOutline()
{

	vrShapeGLCube();
	glColor4ub(0, 0, 0, 255);
#if 0 /* I don't know if the "glDisable(GL_BLEND);" is undesireable or not */
	glBegin(GL_LINE_STRIP);
		glVertex3fv(cube_vertices[0]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[3]);
		glVertex3fv(cube_vertices[0]);
		glVertex3fv(cube_vertices[7]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[7]);
		glVertex3fv(cube_vertices[4]);
		glVertex3fv(cube_vertices[3]);
		glVertex3fv(cube_vertices[4]);
		glVertex3fv(cube_vertices[5]);
	glEnd();
#else
	vrShapeGLCubeOutline();
#endif
}


/*********************************************************************/
void vrShapeGLPyramid()
{
	int	poly;

	for (poly = 0; poly < 5; poly++) {
		glBegin(GL_POLYGON);
			glNormal3fv(pyramid_normals[poly]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][0]]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][1]]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][2]]);
			if (poly == 4)
				glVertex3fv(cube_vertices[pyramid_polygons[poly][3]]);
		glEnd();
	}
}


/*********************************************************************/
void vrShapeGLPyramidOutline()
{
	glPushAttrib(GL_COLOR_BUFFER_BIT);
	glDisable(GL_BLEND);
	glBegin(GL_LINE_STRIP);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[8]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[8]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[6]);
	glEnd();
	glPopAttrib();
}


/*********************************************************************/
void vrShapeGLPyramidPlusOutline()
{

	vrShapeGLPyramid();
	glColor3ub(0, 0, 0);
	vrShapeGLPyramidOutline();
}


/*******************************************************************/
/* vrShapeGLFloor(): draws a floor especially designed for a 10x10 */
/*   operating area.                                               */
/*******************************************************************/
void vrShapeGLFloor()
{
	glColor3ub(25, 25, 200);
	glBegin(GL_POLYGON);
		glVertex3f(-5.0, 0.0, -5.0);
		glVertex3f( 5.0, 0.0, -5.0);
		glVertex3f( 5.0, 0.0,  5.0);
		glVertex3f(-5.0, 0.0,  5.0);
	glEnd();
}


/***************************************************************************/
/* vrGLReportAttributes(): Report the current graphics parameters related  */
/*   to the current color and blending and alpha constructs.               */
/* TODO: add values for glColorMaterial(GL_FRONT, GL_DIFFUSE) & glEnable(GL_COLOR_MATERIAL) */
/* TODO: there are several more texture values that can be reported */
void vrGLReportAttributes(char *msg, unsigned long reports)
{
static	int	header_count = 0;
	int	header_repeat = 30;
	int	report_color = reports & 0x01,
		report_blend = reports & 0x02,
		report_logic = reports & 0x04,
		report_alpha = reports & 0x08,
		report_lights = reports & 0x10,
		report_polys = reports & 0x20,
		report_texture = reports & 0x40,
		report_client_state = reports & 0x80;
	int	rgba_mode;
	float	color[4];
	int	blend_en,
		blend_dst,
		blend_src,
		blend_equ;
	float	blend_clr[4];
	int	logic_mode,
		logic_clr,
		logic_ind;
	int	alpha_en,
		alpha_func;
	float	alpha_ref,
		alpha_bias,
		alpha_scale;
	int	light_en,
		light0_en,
		light1_en,
		light2_en,
		light3_en,
		light4_en,
		light5_en,
		light6_en,
		light7_en,
		light_viewer,
		light_2side;
	float	light_ambclr[4];
	int	poly_cull,
		poly_side,
		poly_front,
		poly_modes[2],
		poly_stipple,
		poly_smooth;
	int	tex_en[3];		/* GL_TEXTURE_1D, GL_TEXTURE_2D, GL_TEXTURE_3D_EXT */
	int	tex_bind[3];
	int	tex_gen[4];		/* GL_TEXTURE_GEN_Q, R, S, T */
	int	tex_map1coords[4];	/* GL_MAP1_TEXTURE_COORD_1, etc */
	int	tex_map2coords[4];	/* GL_MAP2_TEXTURE_COORD_1, etc */
					/* also perhaps GL_POST_TEXTURE_FILTER_BIAS_RANGE_SGIX */
					/* also perhaps GL_POST_TEXTURE_FILTER_SCALE_RANGE_SGIX */
					/* and so on, there are several more texture values */
	int	cstate_colorarray,
		cstate_eflagarray,
		cstate_indexarray,
		cstate_normalarray,
		cstate_tcoordarray,
		cstate_vertexarray;

	/* display nothing if nothing to report */
	if (reports == 0)
		return;

	/*************************/
	/* get the current color */
	if (report_color) {
		glGetIntegerv(GL_RGBA_MODE, &rgba_mode);
		glGetFloatv(GL_CURRENT_COLOR, color);

	/* TODO: add values for glColorMaterial(GL_FRONT, GL_DIFFUSE) & glEnable(GL_COLOR_MATERIAL) */
	}

	/***************************/
	/* get the blending values */
	if (report_blend) {
		glGetIntegerv(GL_BLEND, &blend_en);
		glGetIntegerv(GL_BLEND_DST, &blend_dst);
		glGetIntegerv(GL_BLEND_SRC, &blend_src);
		glGetIntegerv(GL_BLEND_EQUATION_EXT, &blend_equ);
#if defined(GL_BLEND_COLOR_EXT)
                /* TODO: this needs to be improved... JS */
		glGetFloatv(GL_BLEND_COLOR_EXT, blend_clr);
#endif
	}

	/****************************************/
	/* get the pixel logic operation values */
	if (report_logic) {
		glGetIntegerv(GL_LOGIC_OP_MODE, &logic_mode);
		logic_clr = glIsEnabled(GL_COLOR_LOGIC_OP);
		logic_ind = glIsEnabled(GL_INDEX_LOGIC_OP);
	}

	/**********************************/
	/* get the alpha operation values */
	if (report_alpha) {
		alpha_en = glIsEnabled(GL_ALPHA_TEST);
		glGetIntegerv(GL_ALPHA_TEST_FUNC, &alpha_func);
		glGetFloatv(GL_ALPHA_TEST_REF, &alpha_ref);
		glGetFloatv(GL_ALPHA_BIAS, &alpha_bias);
		glGetFloatv(GL_ALPHA_SCALE, &alpha_scale);
	}

	/*************************************/
	/* get the lighting operation values */
	if (report_lights) {
		light_en = glIsEnabled(GL_LIGHTING);
		light0_en = glIsEnabled(GL_LIGHT0);
		light1_en = glIsEnabled(GL_LIGHT1);
		light2_en = glIsEnabled(GL_LIGHT2);
		light3_en = glIsEnabled(GL_LIGHT3);
		light4_en = glIsEnabled(GL_LIGHT4);
		light5_en = glIsEnabled(GL_LIGHT5);
		light6_en = glIsEnabled(GL_LIGHT6);
		light7_en = glIsEnabled(GL_LIGHT7);
		glGetFloatv(GL_LIGHT_MODEL_AMBIENT, light_ambclr);
		glGetIntegerv(GL_LIGHT_MODEL_LOCAL_VIEWER, &light_viewer);
		glGetIntegerv(GL_LIGHT_MODEL_TWO_SIDE, &light_2side);
	}

	/************************************/
	/* get the polygon rendering values */
	if (report_polys) {
		poly_cull = glIsEnabled(GL_CULL_FACE);
		glGetIntegerv(GL_CULL_FACE_MODE, &poly_side);	/* == GL_BACK, GL_FRONT, GL_FRONT_AND_BACK */
		glGetIntegerv(GL_FRONT_FACE, &poly_front);	/* == GL_CCW, GL_CW */
		glGetIntegerv(GL_POLYGON_MODE, poly_modes);	/* == GL_FILL, GL_POINT, GL_LINE*/
		poly_stipple = glIsEnabled(GL_POLYGON_STIPPLE);
		poly_smooth = glIsEnabled(GL_POLYGON_SMOOTH);
	}

	/********************************/
	/* get the current texture info */
	if (report_texture) {
		tex_en[0] = glIsEnabled(GL_TEXTURE_1D);
		tex_en[1] = glIsEnabled(GL_TEXTURE_2D);
#if defined(GL_EXT_texture3D)
		tex_en[2] = glIsEnabled(GL_TEXTURE_3D_EXT);
#else
		tex_en[2] = GL_FALSE;
#endif


#if defined(GL_VERSION_1_1) || defined(GL_VERSION_1_2)
		glGetIntegerv(GL_TEXTURE_BINDING_1D, &(tex_bind[0]));
		glGetIntegerv(GL_TEXTURE_BINDING_2D, &(tex_bind[1]));
#  if 0
                /* TODO: still need to implement a better check or fallback
                 *       for 3D texture stuff - JS
                 */
		glGetIntegerv(GL_TEXTURE_BINDING_3D, &(tex_bind[2]));
#  else
                tex_bind[2] = GL_FALSE;
#  endif

#elif defined(GL_EXT_texture_object)
		glGetIntegerv(GL_TEXTURE_1D_BINDING_EXT, &(tex_bind[0]));
		glGetIntegerv(GL_TEXTURE_2D_BINDING_EXT, &(tex_bind[1]));
#  if defined(GL_TEXTURE_3D_BINDING_EXT)
		glGetIntegerv(GL_TEXTURE_3D_BINDING_EXT, &(tex_bind[2]));
#  else
                tex_bind[2] = GL_FALSE;
#  endif

#else
                tex_bind[0] = GL_FALSE;
                tex_bind[1] = GL_FALSE;
                tex_bind[2] = GL_FALSE;
#endif

		tex_gen[0] = glIsEnabled(GL_TEXTURE_GEN_Q);
		tex_gen[1] = glIsEnabled(GL_TEXTURE_GEN_R);
		tex_gen[2] = glIsEnabled(GL_TEXTURE_GEN_S);
		tex_gen[3] = glIsEnabled(GL_TEXTURE_GEN_T);

		tex_map1coords[0] = glIsEnabled(GL_MAP1_TEXTURE_COORD_1);
		tex_map1coords[1] = glIsEnabled(GL_MAP1_TEXTURE_COORD_2);
		tex_map1coords[2] = glIsEnabled(GL_MAP1_TEXTURE_COORD_3);
		tex_map1coords[3] = glIsEnabled(GL_MAP1_TEXTURE_COORD_4);

		tex_map2coords[0] = glIsEnabled(GL_MAP2_TEXTURE_COORD_1);
		tex_map2coords[1] = glIsEnabled(GL_MAP2_TEXTURE_COORD_2);
		tex_map2coords[2] = glIsEnabled(GL_MAP2_TEXTURE_COORD_3);
		tex_map2coords[3] = glIsEnabled(GL_MAP2_TEXTURE_COORD_4);
	}

	/********************************/
	/* get the current texture info */
	if (report_client_state) {
		cstate_colorarray = glIsEnabled(GL_COLOR_ARRAY);
		cstate_eflagarray = glIsEnabled(GL_EDGE_FLAG_ARRAY);
		cstate_indexarray = glIsEnabled(GL_INDEX_ARRAY);
		cstate_normalarray = glIsEnabled(GL_NORMAL_ARRAY);
		cstate_tcoordarray = glIsEnabled(GL_TEXTURE_COORD_ARRAY);
		cstate_vertexarray = glIsEnabled(GL_VERTEX_ARRAY);
	}


	/********************/
	/* print the header */
	/********************/
	if (header_count == 0) {
		vrPrintf("Message  |");

		if (report_color) {
			vrPrintf("|cmode|color (0-99)|");
		}

		if (report_blend) {
			vrPrintf("|blend|src|dst| equ |color (0-99)|");
		}

		if (report_logic) {
			vrPrintf("|logclr|logind|lmode|");
		}

		if (report_alpha) {
			vrPrintf("|alpha|afunc|a-ref|bias |scale|");
		}

		if (report_lights) {
			vrPrintf("|light| lights |lview|sides|ambclr(0-99)|");
		}

		if (report_polys) {
			vrPrintf("|pcull|faces|dir|rmodes(f,b)|pstip|psmth|");
		}

		if (report_texture) {
			vrPrintf("| tex |binding(1D, 2D, 3D)| gen  | map1c| map2c|");
		}

		if (report_client_state) {
			vrPrintf("| client state ... |");
		}

		vrPrintf("\n");
	}
	header_count++;
	header_count %= header_repeat;

	/********************/
	/* print the values */
	vrPrintf("%9s|", msg);

	if (report_color) {
		vrPrintf("|%s|%3d%3d%3d%3d|",
			(rgba_mode == GL_FALSE ? "index" : "rgba "),
			(int)(color[0] * 99.0), (int)(color[1] * 99.0), (int)(color[2] * 99.0), (int)(color[3] * 99.0));
	}

	if (report_blend) {
		vrPrintf("| %s |%3d|%3d|%5d|%3d%3d%3d%3d|",
			(blend_en == GL_FALSE ? "dis" : "en "),
			blend_src,
			blend_dst,
			blend_equ,
			(int)(blend_clr[0] * 99.0), (int)(blend_clr[1] * 99.0), (int)(blend_clr[2] * 99.0), (int)(blend_clr[3] * 99.0));
	}

	if (report_logic) {
		vrPrintf("| %s  | %s  |%4d |",
			(logic_clr == GL_FALSE ? "dis" : "en "),
			(logic_ind == GL_FALSE ? "dis" : "en "),
			logic_mode);
	}

	if (report_alpha) {
		vrPrintf("| %s |%5d|%5.1f|%5.1f|%5.1f|",
			(alpha_en == GL_FALSE ? "dis" : "en "),
			alpha_func,
			alpha_ref,
			alpha_bias,
			alpha_scale);
	}

	if (report_lights) {
		vrPrintf("| %s |%1d%1d%1d%1d%1d%1d%1d%1d|%4d |%4d |%3d%3d%3d%3d|",
			(light_en == GL_FALSE ? "dis" : "en "),
			(light0_en == GL_FALSE ? 0 : 1),
			(light1_en == GL_FALSE ? 0 : 1),
			(light2_en == GL_FALSE ? 0 : 1),
			(light3_en == GL_FALSE ? 0 : 1),
			(light4_en == GL_FALSE ? 0 : 1),
			(light5_en == GL_FALSE ? 0 : 1),
			(light6_en == GL_FALSE ? 0 : 1),
			(light7_en == GL_FALSE ? 0 : 1),
			light_viewer,
			light_2side,
			(int)(light_ambclr[0] * 99.0), (int)(light_ambclr[1] * 99.0), (int)(light_ambclr[2] * 99.0), (int)(light_ambclr[3] * 99.0));
	}

	if (report_polys) {
		vrPrintf("| %s |%5s|%3s|%5s,%5s| %s | %s |",
			(poly_cull == GL_FALSE ? "dis" : "en "),
			(poly_side == GL_BACK ? "back " :
				(poly_side == GL_FRONT ? "front" :
				(poly_side == GL_FRONT_AND_BACK ? " all " : "none"))),
			(poly_front == GL_CCW ? "ccw" : "cw "),
			(poly_modes[0] == GL_FILL ? "fill " :
				(poly_modes[0] == GL_POINT ? "point" :
				(poly_modes[0] == GL_LINE ? "line " : "???"))),
			(poly_modes[1] == GL_FILL ? "fill " :
				(poly_modes[1] == GL_POINT ? "point" :
				(poly_modes[1] == GL_LINE ? "line " : "???"))),
			(poly_stipple == GL_FALSE ? "dis" : "en "),
			(poly_smooth == GL_FALSE ? "dis" : "en "));
	}

	if (report_texture) {
		vrPrintf("| %c%c%c | %5d %5d %5d | %1d%1d%1d%1d | %1d%1d%1d%1d | %1d%1d%1d%1d |",
			(tex_en[0] == GL_FALSE ? 'd' : 'e'),
			(tex_en[1] == GL_FALSE ? 'd' : 'e'),
			(tex_en[2] == GL_FALSE ? 'd' : 'e'),
			tex_bind[0],
			tex_bind[1],
			tex_bind[2],
			tex_gen[0],
			tex_gen[1],
			tex_gen[2],
			tex_gen[3],
			tex_map1coords[0],
			tex_map1coords[1],
			tex_map1coords[2],
			tex_map1coords[3],
			tex_map2coords[0],
			tex_map2coords[1],
			tex_map2coords[2],
			tex_map2coords[3]);
	}

	if (report_client_state) {
		vrPrintf("| %1d %1d %1d %1d %1d %1d |",
			cstate_colorarray,
			cstate_eflagarray,
			cstate_indexarray,
			cstate_normalarray,
			cstate_tcoordarray,
			cstate_vertexarray
		);
	}

	vrPrintf("\n");
}


/****************************************************************************/
void vrGLRenderDefaultSimulator(vrRenderInfo *renderinfo, int mask)
{
	vrEyeInfo	*curr_eye = renderinfo->eye;
	vrUserInfo	*curr_user = curr_eye->user;
	vrMatrix	dstmat;
	int		num_6sensors = renderinfo->context->input->num_6sensors;
	int		sensor_num;


#define	ATTR_REPORT 0x00	/* Set this value to 0x0 to disable reporting (was 0x3b) */

	/* when the mask is set with all flags off, then no need to render anything */
	if (mask == 0)
		return;

	/* TODO: we should probably either get rid of these, or make them a DBG setting */
	vrGLReportAttributes("Before ", ATTR_REPORT);

	/* Since this routine modifies some of the graphics rendering properties */
	/*   we need to store them here and retrieve them upon exit.             */
	glPushAttrib(0
		| GL_COLOR_BUFFER_BIT		/* covers blending and alpha paramters     */
		| GL_CURRENT_BIT		/* covers current color and rendering mode */
		| GL_POLYGON_BIT		/* covers polygon ordering and style       */
		| GL_LINE_BIT			/* covers line width and style             */
		| GL_LIGHTING_BIT		/* covers all aspects of lighting          */
		| GL_ENABLE_BIT			/* covers all enable bits (inc. texture 2d)*/
	);

#if 1 /* Choose whether to disable lighting, or set to vertex colors as material */
	/* disable lighting and textures */
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);
	/* TODO: add 3D, after checking for that extension */
#else
	/* When lighting is enabled, these two lines cause the */
	/*   vertex colors to be used (otherwise all is grey). */
	/* TODO: should we specifically enable lighting for this style? */
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);
#endif

	/* setup the proper depth checks */
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);

	/* the basicgfx functions assume CCW polygon ordering */
	glFrontFace(GL_CCW);
	glCullFace(GL_BACK);


	/**********************************************/
	/* draw a brown 1' cube to represent the head */
	glPushMatrix(); /* { */
	glMultMatrixd(vrMatrixGet6sensorValuesDirect(&dstmat, curr_user->head)->v);

	/* the "head" */
	glColor3fv(curr_user->color);
	vrShapeGLCube();

	/* use the head's outline to signify the sensor status */
	if (curr_user->head->error > 0)
		glColor3ub((GLbyte)230, (GLbyte)0, (GLbyte)0);	/* red   -- error */
	else if (curr_user->head->oob)
		glColor3ub((GLbyte)0, (GLbyte)0, (GLbyte)230);	/* blue  -- out of bounds */
	else if (curr_user->head->active)
		glColor3ub((GLbyte)0, (GLbyte)230, (GLbyte)0);	/* green -- active */
	else
		glColor3ub((GLbyte)0, (GLbyte)0, (GLbyte)0);	/* black -- inactive, in bounds, no error */
	vrShapeGLCubeOutline();

	/* the "nose" */
	glPushMatrix(); /* { */
	glTranslatef(0.0, 0.0, -1.0);
	glScalef(0.1, 0.2, 0.5);
	glColor3f(curr_user->color[0]*0.5, curr_user->color[1]*0.5, curr_user->color[2]*0.5);
	vrShapeGLCubePlusOutline();
	glPopMatrix(); /* } */

	/* TODO: the eye positions should use the actual values. */
	/* the "left eye" */
	glPushMatrix(); /* { */
	glTranslatef(-0.5, 0.5, -1.0);
	glScalef(0.25, 0.25, 0.25);
	glColor3ub(255, 255, 255);
	vrShapeGLCubePlusOutline();
	glPopMatrix(); /* } */

	/* the "right eye" */
	glPushMatrix(); /* { */
	glTranslatef(0.5, 0.5, -1.0);
	glScalef(0.25, 0.25, 0.25);
	glColor3ub(255, 255, 255);
	vrShapeGLCubePlusOutline();
	glPopMatrix(); /* } */

	glPopMatrix(); /* } */


	/**************************************************************/
	/* draw a simple object to represent sensor-1 (usu. the wand) */
#if 1 /* compile-time option: draw the wand? { */
	if (num_6sensors >= 2) {
		glPushMatrix(); /* { */
		glMultMatrixd(vrMatrixGet6sensorValues(&dstmat, 1)->v);

		/* the "wand" */
#  if 1 /* draw as a pyramid */
		glColor3fv(curr_user->color);
		glScalef(0.20, 0.20, 0.75);
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		vrShapeGLPyramid();

		/* use the sensor representation outline to signify the sensor status */
		if (vrGet6sensorErrorValue(1) > 0)
			glColor3ub((GLbyte)230, (GLbyte)0, (GLbyte)0);	/* red   -- error */
		else if (vrGet6sensorOobValue(1))
			glColor3ub((GLbyte)0, (GLbyte)0, (GLbyte)230);	/* blue  -- out of bounds */
		else if (vrGet6sensorActiveValue(1))
			glColor3ub((GLbyte)0, (GLbyte)230, (GLbyte)0);	/* green -- active */
		else
			glColor3ub((GLbyte)0, (GLbyte)0, (GLbyte)0);	/* black -- inactive, in bounds, no error */
		vrShapeGLPyramidOutline();

#  else /* line-method */
		{
		vrPoint		proploc;
		vrPoint		drawto;
		vrVector	vec;

		vrPointGetRWFrom6sensor(&proploc, 1);
		vrVectorGetRWFrom6sensorDir(&vec, 1, VRDIR_FORE);
		vrPointAddScaledVector(&drawto, &proploc, &vec, 2.0);
		glColor4ub(255, 255, 0, 255);
		glDisable(GL_BLEND);
		glBegin(GL_LINE_STRIP);
			glVertex3dv(proploc.v);
			glVertex3dv(drawto.v);
		glEnd();
		}
#  endif
		glPopMatrix(); /* } */
	}

#  if 1 /* compile-time option: draw the other sensors as boxes? */
	/* draw simple objects to represent sensors 2-n */
	for (sensor_num = 2; sensor_num < num_6sensors; sensor_num++) {
		glPushMatrix(); /* { */
		glMultMatrixd(vrMatrixGet6sensorValues(&dstmat, sensor_num)->v);

		glColor3fv(curr_user->color);
		glScalef(0.20, 0.20, 0.20);
		vrShapeGLCube();

		/* use the sensor representation outline to signify the sensor status */
		if (vrGet6sensorErrorValue(sensor_num) > 0)
			glColor3ub((GLbyte)230, (GLbyte)0, (GLbyte)0);	/* red   -- error */
		else if (vrGet6sensorOobValue(sensor_num))
			glColor3ub((GLbyte)0, (GLbyte)0, (GLbyte)230);	/* blue  -- out of bounds */
		else if (vrGet6sensorActiveValue(sensor_num))
			glColor3ub((GLbyte)0, (GLbyte)230, (GLbyte)0);	/* green -- active */
		else
			glColor3ub((GLbyte)0, (GLbyte)0, (GLbyte)0);	/* black -- inactive, in bounds, no error */
		vrShapeGLCubeOutline();

		glPopMatrix(); /* } */
	}
#  endif
#endif /* } draw wand */


	/******************************************************************/
	/* draw the outline and a semi-transparent surface of a 10^3 CAVE */
	glPushMatrix(); /* { */

	glLineWidth(1.5);
	glPolygonMode(GL_FRONT, GL_POINT);

	/* move and scale how the cube will be drawn */
	/* TODO: it would be nice to base this off some information stored */
	/*   with the window about the size of the working volume.         */
	glTranslatef(0.0, 5.0, 0.0);
	glScalef(4.999, 4.999, 4.999);

	/* draw the outline */
	glColor3ub(230, 230, 230);
	vrShapeGLCubeOutline();

#if 1 /* compile-time: draw the semi-transparent CAVE screen surface */
	glDisable(GL_CULL_FACE);	/* since we want to see the screens from inside or out */
	/* TODO: figure out why the "outside" of the CAVE screen box isn't rendered */

	/* draw the surface */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	/* also works with GL_ONE as arg2 */
	glColor4ub(200, 200, 200, 64);
	vrShapeGLCube();

	/* redraw the outline to cover holes made by screen surface */
	glColor3ub(230, 230, 230);
	vrShapeGLCubeOutline();

	glColor4ub(255, 255, 255, 255);
#endif /* draw cave surface vs. just the white outline */

	vrGLReportAttributes("Middle ", ATTR_REPORT);

	glPopMatrix(); /* } */

	glPopAttrib();
	vrGLReportAttributes("After  ", ATTR_REPORT);
}

